# Copyright 2014 Florian Ludwig
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import, division, print_function, with_statement

import rw.gen
import rw.event
import rw.scope


class Plugin(object):
    def __init__(self, name):
        self.name = name
        self.activate_event = rw.event.Event('PLUGIN_{}_ACTIVATE'.format(name))

    def init(self, function):
        function = rw.scope.inject(function)
        self.activate_event.add(function)
        return function

    @rw.scope.inject
    @rw.gen.coroutine
    def activate(self, scope):
        plugins = scope.get('rw.plugins', {})
        if self not in plugins:
            yield self.activate_event()
            plugins[self] = {}